"""Logger."""
import logging
import sys

log_format = logging.Formatter("[%(asctime)s] %(levelname)-8s %(name)-17s %(message)s")
log_level = logging.DEBUG
logger = logging.getLogger("t_bug_catcher")
if logger.hasHandlers():
    logger.handlers = []
logger.setLevel(log_level)
logger.propagate = False

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(log_level)
handler.setFormatter(log_format)
logger.addHandler(handler)



import sys
from typing import Optional

import bugsnag
import requests

from .config import CONFIG
from .utils import logger
from .workitems import variables

bugsnag.configuration.auto_notify = False


class BugSnag:
    """BugSnag class for interacting with the BugSnag API."""

    def __init__(self):
        """Initializes the BugSnag class."""
        pass

    def config(self, api_key: str) -> bool:
        """Configures the BugSnag class.

        Args:
            api_key (str): The API key for the BugSnag account.

        Returns:
            bool: True if the configuration was successful, False otherwise.
        """
        try:
            bugsnag.configure(api_key=api_key, release_stage=CONFIG.ENVIRONMENT, auto_notify=False)
            bugsnag.add_metadata_tab(
                "Metadata",
                {
                    "run_url": variables.get("processRunUrl", ""),
                    "run_by": variables.get("userEmail", ""),
                },
            )
            response = requests.request(
                "POST",
                "https://otlp.bugsnag.com/v1/traces",
                headers={
                    "Content-Type": "application/json",
                    "Bugsnag-Api-Key": api_key,
                    "Bugsnag-Payload-Version": "4",
                    "Bugsnag-Span-Sampling": "True",
                },
                data='{"message": "test"}',
            )
            if response.status_code not in [200, 201, 202, 204]:
                logger.warning(f"Error connecting to Bugsnag: {response.text}")
                return False
            return True
        except Exception as ex:
            logger.warning(f"Failed to configure Bugsnag: {ex}")
            return False

    def report_error(self, exception: Optional[Exception] = None, metadata: Optional[dict] = None):
        """Sends an error to BugSnag.

        Args:
            exception (Exception, optional): The exception to report.
            metadata (dict, optional): The metadata to be added to the Bugsnag issue. Defaults to None.

        Returns:
            None
        """
        if not exception:
            _, exception, _ = sys.exc_info()
        if isinstance(metadata, dict):
            bugsnag.notify(exception=exception, metadata={"special_info": metadata})
            return
        if metadata is None:
            bugsnag.notify(exception=exception)
            return
        logger.warning(f"Incorrect type of metadata: {type(metadata)}")
        bugsnag.notify(exception=exception)

    def report_unhandled_error(self, exc_type, exc_value, traceback):
        """Sends an unhandled exception to BugSnag.

        Args:
            exc_type (type): The type of the exception.
            exc_value (Exception): The value of the exception.
            traceback (traceback): The traceback of the exception.

        Returns:
            None
        """
        bugsnag.notify((exc_type, exc_value, traceback), severity="error")



import json
import linecache
import re
import sys
from datetime import datetime
from types import FunctionType, ModuleType
from typing import Optional

from .config import CONFIG
from .utils import logger
from .utils.common import Encoder, convert_keys_to_primitives, strip_path


class StackSaver:
    """A class to save the stack trace."""

    def __init__(self):
        """Initializes the StackSaver class."""
        pass

    def serialize_frame_info(self, frame_info: dict) -> dict:
        """A static method to serialize the frame info.

        Args:
            frame_info (dict): The frame info to be serialized.

        Returns:
            dict: The serialized frame info.
        """
        run_locals = convert_keys_to_primitives(frame_info["locals"]) if frame_info["locals"] else {}
        serializable_frame_info = {
            "filename": frame_info["filename"],
            "function_name": frame_info["function_name"],
            "line_number": frame_info["line_number"],
            "line": frame_info["line"],
            "locals": run_locals,
        }
        return serializable_frame_info

    @staticmethod
    def filter_variables(variables: dict) -> dict:
        """A static method to filter the variables.

        Args:
            variables (dict): The variables to be filtered.

        Returns:
            dict: The filtered variables.
        """
        if not isinstance(variables, dict):
            return variables
        else:
            local_variables = {}
            for var_name, var in variables.items():
                if re.match(r"^__\w+__$", var_name):
                    continue
                if isinstance(var, (ModuleType, FunctionType)):
                    continue
                local_variables[str(var_name)] = var
            return local_variables

    def mask_credentials(self, file_path: str) -> None:
        """A method to mask the credentials in the file.

        Args:
            file_path (str): The path of the file to be masked.

        Raises:
            Exception: If the masking fails.

        Returns:
            None
        """
        partial_match_pattern = re.compile(
            r"(['\"])([^'\"]*?(%s)[^'\"]*?)(['\"])\s*([:,])\s*(['\"])([^'\"]*)(['\"])"
            % (r"|".join(CONFIG.PARTIAL_MATCH_KEYS)),
            re.IGNORECASE,
        )

        full_match_pattern = re.compile(
            r"(['\"])(%s)(['\"])\s*([:,])\s*(['\"])([^'\"]*)(['\"])" % (r"|".join(CONFIG.SENSITIVE_KEYS)), re.IGNORECASE
        )

        with open(file_path, "r") as f:
            filedata = f.read()

        partial_replace = self.mask_data((1, 2, 5))
        full_replace = self.mask_data((1, 2, 4))

        masked_data = partial_match_pattern.sub(partial_replace, filedata)
        masked_data = full_match_pattern.sub(full_replace, masked_data)

        with open(file_path, "w") as file:
            file.write(masked_data)

    def mask_data(self, indicates: tuple):
        """A method to mask the data.

        Args:
            indicates (tuple): The indicates of the data to be masked.

        Returns:
            function: The function to mask the data.
        """

        def mask_sensitive(match):
            if match.group(2):
                quote = match.group(indicates[0])
                key_part = match.group(indicates[1])
                delimiter = match.group(indicates[2])
                return f"{quote}{key_part}{quote}{delimiter} {quote}***{quote}"

        return mask_sensitive

    def save_stack_trace(self, exception: Optional[Exception] = None):
        """A method to save the stack trace.

        Args:
            exception (Exception, optional): The exception to be saved. Defaults to None.

        Returns:
            Optional[str]: The path of the saved stack trace.
        """
        try:
            frames = []
            stack_details_json = []
            seen_frames = set()
            tb = exception.__traceback__ if exception else sys.exc_info()[2]
            while tb is not None:
                frame = tb.tb_frame
                if frame.f_code.co_name in seen_frames:
                    tb = tb.tb_next
                    continue
                seen_frames.add(frame.f_code.co_name)
                if "site-packages" in frame.f_code.co_filename:
                    tb = tb.tb_next
                    continue
                frames.append(frame)
                tb = tb.tb_next
            frames = frames[-CONFIG.LIMITS.STACK_SCOPE :]

            for frame in frames:
                frame_info = {
                    "filename": strip_path(frame.f_code.co_filename),
                    "function_name": frame.f_code.co_name,
                    "line_number": frame.f_lineno,
                    "line": linecache.getline(frame.f_code.co_filename, frame.f_lineno).strip(),
                    "locals": self.filter_variables(frame.f_locals),
                }
                stack_details_json.append(self.serialize_frame_info(frame_info))

            file_path = f"stack_details_{datetime.now().strftime('%Y%m%d%H%M%S')}.json"

            with open(file_path, "w") as f:
                json.dump(stack_details_json, f, indent=4, cls=Encoder)

            self.mask_credentials(file_path)

            return file_path
        except Exception as e:
            logger.warning(f"Failed to save stack trace: {e}")
            return



import copy
import json
import os
import re
import traceback
import urllib
from datetime import date, datetime
from pathlib import Path
from types import TracebackType
from typing import List

from ..config import CONFIG


class Encoder(json.JSONEncoder):
    """This class is used to encode the Episode object to json."""

    def default(self, o):
        """This method is used to encode the Episode object to json.

        Args:
            o (object): The object to be encoded.

        Returns:
            str: The json string.
        """
        try:
            object_copy = copy.deepcopy(o)
        except (TypeError, AttributeError):
            return str(o)[: CONFIG.LIMITS.STACK_TEXT_LENGTH]
        try:
            if hasattr(object_copy, "__dict__"):
                keys_to_remove = [
                    key
                    for key in object_copy.__dict__.keys()
                    if any(s in str(key).lower() for s in CONFIG.KEYS_TO_REMOVE)
                ]
                for key in keys_to_remove:
                    del object_copy.__dict__[key]
                return {
                    str(key): str(value)[: CONFIG.LIMITS.STACK_TEXT_LENGTH]
                    for key, value in object_copy.__dict__.items()
                }
            if isinstance(object_copy, (datetime, date)):
                return object_copy.isoformat()
            if isinstance(object_copy, Path):
                return str(object_copy)
            return super().default(self, object_copy)
        except TypeError:
            return str(object_copy)[: CONFIG.LIMITS.STACK_TEXT_LENGTH]


def get_frames(exc_traceback: TracebackType) -> List:
    """Get the frames of the exception.

    Args:
        exc_traceback (TracebackType): The traceback of the exception.

    Returns:
        List: The frames of the exception.
    """
    frames = [
        frame for frame in traceback.extract_tb(exc_traceback) if "site-packages" not in str(frame.filename).lower()
    ]
    if not frames:
        frames = [frame for frame in traceback.extract_tb(exc_traceback)]
    return frames


def convert_keys_to_primitives(data: dict) -> dict:
    """A function that recursively converts keys in a nested dictionary to primitives.

    Args:
        data (dict): The input dictionary to convert keys.

    Returns:
        dict: A new dictionary with keys converted to strings.
    """
    new_dict = {}
    for key, value in data.items():
        if isinstance(value, list) or isinstance(value, tuple):
            value = value[: CONFIG.LIMITS.STACK_ITEM_LENGTH]
        if isinstance(value, dict):
            items_list = list(value.items())
            sliced_list = items_list[: CONFIG.LIMITS.STACK_ITEM_LENGTH]
            value = dict(sliced_list)
            new_dict[str(key)] = convert_keys_to_primitives(value)
        else:
            new_dict[str(key)] = value
    return new_dict


def strip_path(path: str):
    """A function to strip the current working directory path from the input.

    Args:
        path (str): The path from which to strip the current working directory path.

    Returns:
        str: The stripped path.
    """
    return path.replace(os.getcwd(), "").strip(os.sep)


def remove_holotree_id(path: str) -> str:
    """A function to remove the value after 'holotree' from the path.

    Args:
        path (str): The path from which to remove the value after 'holotree'.

    Returns:
        str: The modified path.
    """
    pattern = r"(holotree\/)[^\/]+\/"
    return re.sub(pattern, r"\1", path)


def clean_nested_data(data):
    """Recursively replaces None values with empty strings in nested dictionaries and lists.

    Args:
        data: Input data structure (dict, list, or other type)

    Returns:
        Cleaned data structure with None values replaced by empty strings
    """
    if isinstance(data, dict):
        return {key: clean_nested_data(value) for key, value in data.items()}
    elif isinstance(data, (list, tuple)):
        return [clean_nested_data(item) for item in data]
    else:
        return "" if data is None else data


def retrieve_build_info():
    """Logs build information."""
    if not os.path.exists(CONFIG.BUILD_INFO_FILE):
        return
    with open(CONFIG.BUILD_INFO_FILE, "r") as json_file:
        commit_info = json.load(json_file)
    repository_url = commit_info["repository_url"]
    current_branch = commit_info["branch"]
    encoded_branch = urllib.parse.quote(current_branch)
    commit_info["branch_url"] = f"{repository_url}/src/{commit_info['last_commit']}/?at={encoded_branch}"
    commit_info["commit_url"] = f"{repository_url}/commits/{commit_info['last_commit']}/"
    return commit_info



try:
    from RPA.Robocorp.WorkItems import WorkItems

    work_items = WorkItems()
    work_items.get_input_work_item()
    work_item = work_items.get_work_item_variables()
    variables = work_item.get("variables", dict())
    metadata = work_item.get("metadata", dict())
except (ImportError, KeyError):
    variables = {}
    metadata = {}



import os
import unittest
from unittest.mock import PropertyMock, patch

from t_bug_catcher.bug_catcher import BugCatcher
from t_bug_catcher.config import CONFIG
from t_bug_catcher.utils import logger


class TestBugCatcher(unittest.TestCase):
    """TestBugCatcher class."""

    def setUp(self):
        """Set up."""
        self.bug_catcher = BugCatcher()

    def test_report_error_with_local_environment(self):
        """Test report_error function."""
        with patch.object(logger, "warning") as mock_warning:
            self.bug_catcher.report_error()
            mock_warning.assert_called_once_with("Reporting an error is not supported in local environment.")

    @patch.object(CONFIG, "ENVIRONMENT", "robocloud")
    def test_report_error_with_no_configurations(self):
        """Test report_error function."""
        with patch.object(logger, "warning") as mock_warning:
            self.bug_catcher.report_error()
            mock_warning.assert_called_once_with(
                "Jira and BugSnag are not configured. Please configure them before reporting an error."
            )

    def test_jira_configure(self):
        """Test jira_configure function."""
        self.bug_catcher.configure.jira(
            login=os.getenv("JIRA_LOGIN"),
            api_token=os.getenv("JIRA_API_TOKEN"),
            project_key=os.getenv("JIRA_PROJECT_KEY"),
        )
        assert self.bug_catcher.configure.is_jira_configured

    def test_bugsnag_configure(self):
        """Test bugsnag_configure function."""
        self.bug_catcher.configure.bugsnag(api_key=os.getenv("BUGSNAG_API_KEY"))
        assert self.bug_catcher.configure.is_bugsnag_configured

    def test_attach_file(self):
        """Test attach_file function."""
        ex = Exception("some error")
        self.bug_catcher.attach_file_to_exception(exception=ex, attachment="test.txt")
        assert hasattr(ex, "custom_attachments")

    @patch.object(CONFIG, "ENVIRONMENT", "robocloud")
    def test_report_error(self):
        """Test report_error function."""
        with patch.object(
            self.bug_catcher, "_BugCatcher__configurator", new_callable=PropertyMock
        ) as mock_configurator:
            patch.object(mock_configurator, "is_jira_configured", return_value=True)
            patch.object(mock_configurator, "is_bugsnag_configured", return_value=True)
            try:
                raise Exception("Test exception")
            except Exception as ex:
                with patch.object(logger, "warning") as mock_warning:
                    self.bug_catcher.report_error(exception=ex)
                    actual_call = mock_warning.call_args
                    warning_message = actual_call[0][0]
                    self.assertTrue(warning_message.startswith("Failed to report Bug Catcher issue due to"))
                    self.assertEqual(mock_warning.call_count, 2)


if __name__ == "__main__":
    unittest.main()



import os
import socket

from .workitems import metadata, variables


class Config:
    """Config class for configuring the application."""

    class LIMITS:
        """Limits class for configuring the application."""

        MAX_ATTACHMENTS: int = 5
        MAX_ISSUE_ATTACHMENTS: int = 100
        MAX_DESCRIPTION_LENGTH: int = 250
        SUMMARY_LENGTH: int = 120
        STACK_SCOPE: int = 3
        STACK_ITEM_LENGTH: int = 100
        STACK_TEXT_LENGTH: int = 10000
        COMMENT_LIMIT: int = 4950

    class TICKET_PRIORITIES:
        """Priorities class for configuring the application."""

        HIGHEST: str = "1"
        HIGH: str = "2"
        MEDIUM: str = "3"
        LOW: str = "4"
        LOWEST: str = "5"

    SUPPORT_BOARD = "AB"
    BC_BOARD = "BC"

    KEYS_TO_REMOVE = ["credential", "password"]

    SENSITIVE_KEYS = [
        r"user(?:name|_name)?",
        r"id",
        r"uid",
        r"otp",
        r"api(?:[_]?key|[_]?token)",
        r"\w+api(?:[_]?key|[_]?token)",
        r"key",
        r"account",
        r"client_(?:id|secret)",
    ]
    PARTIAL_MATCH_KEYS = [r"login", r"password", r"token", r"secret"]

    BUILD_INFO_FILE = "commit_info.json"

    HOST = socket.gethostname() if os.name.lower() == "nt" else None
    UNAME = os.getlogin() if os.name.lower() == "nt" else None

    RC_RUN_LINK = (
        f"https://cloud.robocorp.com/organizations/{os.environ.get('RC_ORGANIZATION_ID')}"
        f"/workspaces/{os.environ.get('RC_WORKSPACE_ID')}/processes"
        f"/{os.environ.get('RC_PROCESS_ID')}/runs/{os.environ.get('RC_PROCESS_RUN_ID')}/"
        f"stepRuns/{os.environ.get('RC_ACTIVITY_RUN_ID')}/"
    )

    ENVIRONMENT = (
        "robocloud"
        if not variables.get("environment") and os.environ.get("RC_PROCESS_RUN_ID")
        else variables.get("environment", "local")
    )

    STAGE = metadata.get("process", dict()).get("implementationStage", "")
    ADMIN_CODE = metadata.get("process", dict()).get("adminCode", "")
    WORKER_NAME = metadata.get("process", dict()).get("name", "")
    EMPOWER_URL = metadata.get("process", dict()).get("processRunUrl") or variables.get("processRunUrl")


CONFIG = Config()
